#include "cuda_runtime.h"
#include <cuda.h>
#include <opencv2/opencv.hpp>
#include <stdio.h>  
#include <windows.h>  
#include "iostream"	
#include <io.h>
#include <string>
#include <cstring>
#include <vector>
#include <fstream>

using namespace std;
using namespace cv;

vector<string> files;

void getFiles(string path)
{
	//handle
	long   hFile   =   0;

	//file information
	struct _finddata_t fileinfo;
	string p;
	if((hFile = _findfirst(p.assign(path).append("\\*").c_str(),&fileinfo)) !=  -1)
	{
		do
		{
			if((fileinfo.attrib &  _A_SUBDIR))
			{
				if(strcmp(fileinfo.name,".") != 0  &&  strcmp(fileinfo.name,"..") != 0)
					getFiles( p.assign(path).append("\\").append(fileinfo.name));
			}
			else
			{
				files.push_back(p.assign(path).append("\\").append(fileinfo.name) );
			}
		}while(_findnext(hFile, &fileinfo)  == 0);
		_findclose(hFile);
	}
}

//Sobel Edge Detection by CPU
void SobelInCPU(Mat srcImg, Mat dstImg, int imgHeight, int imgWidth)
{
	int Gx = 0;
	int Gy = 0;
	for (int i = 1; i < imgHeight - 1; i++)
	{
		uchar *dataUp = srcImg.ptr<uchar>(i - 1);
		uchar *data = srcImg.ptr<uchar>(i);
		uchar *dataDown = srcImg.ptr<uchar>(i + 1);
		uchar *out = dstImg.ptr<uchar>(i);
		for (int j = 1; j < imgWidth - 1; j++)
		{
			Gx = (dataUp[j + 1] + 2 * data[j + 1] + dataDown[j + 1]) - (dataUp[j - 1] + 2 * data[j - 1] + dataDown[j - 1]);
			Gy = (dataUp[j - 1] + 2 * dataUp[j] + dataUp[j + 1]) - (dataDown[j - 1] + 2 * dataDown[j] + dataDown[j + 1]);
			out[j] = (abs(Gx) + abs(Gy)) / 2;
		}
	}
}
		  
extern "C" void cSobelInCuda(unsigned char* dstMat, const unsigned char* srcMat, const int srcHeight, const int srcWidth);

void solve()
{

	LARGE_INTEGER _begin;
	LARGE_INTEGER _end;
	LARGE_INTEGER tmp;

	LARGE_INTEGER _begin1;
	LARGE_INTEGER _end1;
	LARGE_INTEGER tmp1;

	long costTimeCPU=0;
	long costTimeGPU=0;

	QueryPerformanceFrequency(&tmp);
	LONGLONG _freq = tmp.QuadPart;
  	QueryPerformanceFrequency(&tmp1);
	LONGLONG _freq1 = tmp1.QuadPart;
		
	int iSizeGray = 0;
	int iSizeColor = 0;
	unsigned char* deviceSrc;
	unsigned char* deviceDst;
												  
	string 	filePath = "..//Image_Simulate_Mem//Input_Images";
	getFiles(filePath);

	int size = files.size();
	int pos = filePath.length();
	for (int i = 0; i < size; i++)
		files[i] == files[i].erase(0, pos + 1);

	ofstream f1("Cost Time.txt");
	if(!f1)
		return;

	string imagefiles = "";

	for (int i = 0; i < size; i++) 
	{

		imagefiles = "Input_Images//";
		imagefiles += files[i];

		cv::Mat img= cv::imread(imagefiles, 0);	  
		cv::Mat dst= cv::Mat(img.size(), CV_8UC1);
		cv::Mat dst1= cv::Mat(img.size(), CV_8UC1);
		iSizeGray= img.cols * img.rows;
		iSizeColor= iSizeGray * 1;

		cudaMalloc(&deviceSrc, iSizeColor);
		cudaMalloc(&deviceDst, iSizeColor);
		cudaMemcpy(deviceSrc, img.data, iSizeColor, cudaMemcpyHostToDevice);

		// CPU time	costs 10 times
		QueryPerformanceCounter(&_begin);

		for (int k = 0; k < 10; k++) 
		{	
			  SobelInCPU(img, dst, img.rows, img.cols);
		}

		QueryPerformanceCounter(&_end);

		QueryPerformanceCounter(&_begin1);

		// Simulate Membrane by GPU (MGPU), time costs 10000 times
		for (int k = 0; k < 10000; k++)
		{
			cSobelInCuda(deviceDst, deviceSrc, img.rows, img.cols);
		}

		QueryPerformanceCounter(&_end1);	
		costTimeGPU = (long)((_end1.QuadPart - _begin1.QuadPart)*1000 / _freq1);
		costTimeCPU = (long)((_end.QuadPart - _begin.QuadPart)*1000 / _freq);

		cudaMemcpy(dst1.data, deviceDst, iSizeColor, cudaMemcpyDeviceToHost);

		f1 << files[i] << endl;
		f1 <<"TimeMGPU  "<<costTimeGPU/10000.0<< endl;
		f1 <<"TimeCPU   "<<costTimeCPU/10.0<<"\n"<< endl;

		string OutPut = "Output_Image//"	;

		string CPUSavename = OutPut + "resultCPU"  +	 files[i];
		string GPUSavename = OutPut + "resultGPU"  +	 files[i];
	
		cv::imwrite(CPUSavename,dst);
		cv::imwrite(GPUSavename,dst1);

	}
	f1.close();
	cv::waitKey(0);
}


int main()
{
	solve();
	return 0;
}